#include <stdlib.h>
#include <string.h>
#include "linear.h"

#include "mex.h"

#ifdef MX_API_VER
#if MX_API_VER < 0x07030000
typedef int mwIndex;
#endif
#endif

#define Malloc(type,n) (type *)malloc((n)*sizeof(type))

#define NUM_OF_RETURN_FIELD 7

static const char *field_names[] = {
	"Parameters",
	"nr_class",
	"nr_feature",
	"bias",
	"Label",
	"w",
	"rho",
};

const char *model_to_matlab_structure(mxArray *plhs[], struct model *model_)
{
	int i, j, k, nnz;
	int nr_w;
	double *ptr;
	mxArray *return_model, **rhs;
	int out_id = 0;
	int n, w_size;
	mwIndex *ir, *jc;

	rhs = (mxArray **)mxMalloc(sizeof(mxArray *)*NUM_OF_RETURN_FIELD);

	// Parameters
	// for now, only solver_type and ovo are needed
	rhs[out_id] = mxCreateDoubleMatrix(1, 2, mxREAL);
	ptr = mxGetPr(rhs[out_id]);
	ptr[0] = model_->param.solver_type;
	ptr[1] = model_->param.ovo;
	out_id++;

	// nr_class
	rhs[out_id] = mxCreateDoubleMatrix(1, 1, mxREAL);
	ptr = mxGetPr(rhs[out_id]);
	ptr[0] = model_->nr_class;
	out_id++;

	if(model_->nr_class==2 && model_->param.solver_type != MCSVM_CS)
		nr_w=1;
	else if(model_->param.ovo)
		nr_w=model_->nr_class*(model_->nr_class-1)/2;
	else
		nr_w=model_->nr_class;

	// nr_feature
	rhs[out_id] = mxCreateDoubleMatrix(1, 1, mxREAL);
	ptr = mxGetPr(rhs[out_id]);
	ptr[0] = model_->nr_feature;
	out_id++;

	// bias
	rhs[out_id] = mxCreateDoubleMatrix(1, 1, mxREAL);
	ptr = mxGetPr(rhs[out_id]);
	ptr[0] = model_->bias;
	out_id++;

	if(model_->bias>=0)
		n=model_->nr_feature+1;
	else
		n=model_->nr_feature;

	w_size = n;
	// Label
	if(model_->label)
	{
		rhs[out_id] = mxCreateDoubleMatrix(model_->nr_class, 1, mxREAL);
		ptr = mxGetPr(rhs[out_id]);
		for(i = 0; i < model_->nr_class; i++)
			ptr[i] = model_->label[i];
	}
	else
		rhs[out_id] = mxCreateDoubleMatrix(0, 0, mxREAL);
	out_id++;

	// w
	if(model_->param.ovo) // ovo
	{
		nnz = 0;
		for(i = 0; i < w_size; i++)
			nnz += model_->sparse_w_nnz[i];

		rhs[out_id] = mxCreateSparse(nr_w, w_size, nnz, mxREAL);
		ptr = mxGetPr(rhs[out_id]);
		ir = mxGetIr(rhs[out_id]);
		jc = mxGetJc(rhs[out_id]);

		k = 0;
		for(i = 0; i < w_size; i++)
		{
			jc[i] = (mwIndex) k;
			for(j = 0; j < model_->sparse_w_nnz[i]; j++)
			{
				ir[k] = (mwIndex) (model_->sparse_w[i]+j)->index-1;
				ptr[k] = (model_->sparse_w[i]+j)->value;
				k++;
			}
		}
		jc[i] = (mwIndex) k;
	}
	else // ovr
	{
		rhs[out_id] = mxCreateDoubleMatrix(nr_w, w_size, mxREAL);
		ptr = mxGetPr(rhs[out_id]);
		for(i = 0; i < w_size*nr_w; i++)
			ptr[i]=model_->w[i];
	}
	out_id++;

	// rho
	rhs[out_id] = mxCreateDoubleMatrix(1, 1, mxREAL);
	ptr = mxGetPr(rhs[out_id]);
	ptr[0] = model_->rho;
	out_id++;

	/* Create a struct matrix contains NUM_OF_RETURN_FIELD fields */
	return_model = mxCreateStructMatrix(1, 1, NUM_OF_RETURN_FIELD, field_names);

	/* Fill struct matrix with input arguments */
	for(i = 0; i < NUM_OF_RETURN_FIELD; i++)
		mxSetField(return_model,0,field_names[i],mxDuplicateArray(rhs[i]));
	/* return */
	plhs[0] = return_model;
	mxFree(rhs);

	return NULL;
}

const char *matlab_matrix_to_model(struct model *model_, const mxArray *matlab_struct)
{
	int i, j, k, num_of_fields, nnz;
	int nr_w;
	double *ptr;
	int id = 0;
	int n, w_size;
	mxArray **rhs;
	mwIndex *ir, *jc;

	num_of_fields = mxGetNumberOfFields(matlab_struct);
	rhs = (mxArray **) mxMalloc(sizeof(mxArray *)*num_of_fields);

	for(i=0;i<num_of_fields;i++)
		rhs[i] = mxGetFieldByNumber(matlab_struct, 0, i);

	model_->nr_class=0;
	nr_w=0;
	model_->nr_feature=0;
	model_->w=NULL;
	model_->sparse_w=NULL;
	model_->sparse_w_nnz=NULL;
	model_->label=NULL;

	// Parameters
	ptr = mxGetPr(rhs[id]);
	model_->param.solver_type = (int)ptr[0];
	model_->param.ovo = (int)ptr[1];
	id++;

	// nr_class
	ptr = mxGetPr(rhs[id]);
	model_->nr_class = (int)ptr[0];
	id++;

	if(model_->nr_class==2 && model_->param.solver_type != MCSVM_CS)
		nr_w=1;
	else if(model_->param.ovo)
		nr_w=model_->nr_class*(model_->nr_class-1)/2;
	else
		nr_w=model_->nr_class;

	// nr_feature
	ptr = mxGetPr(rhs[id]);
	model_->nr_feature = (int)ptr[0];
	id++;

	// bias
	ptr = mxGetPr(rhs[id]);
	model_->bias = ptr[0];
	id++;

	if(model_->bias>=0)
		n=model_->nr_feature+1;
	else
		n=model_->nr_feature;
	w_size = n;

	// Label
	if(mxIsEmpty(rhs[id]) == 0)
	{
		model_->label = Malloc(int, model_->nr_class);
		ptr = mxGetPr(rhs[id]);
		for(i=0;i<model_->nr_class;i++)
			model_->label[i] = (int)ptr[i];
	}
	id++;

	// w
	if(model_->param.ovo) // ovo
	{
		ptr = mxGetPr(rhs[id]);
		ir = mxGetIr(rhs[id]);
		jc = mxGetJc(rhs[id]);
		nnz = jc[w_size];
		model_->sparse_w = Malloc(struct feature_node*, w_size);
		model_->sparse_w_nnz = Malloc(int, w_size);
		struct feature_node **sparse_w = model_->sparse_w;
		int *sparse_w_nnz = model_->sparse_w_nnz;

		struct feature_node *w_space = Malloc(struct feature_node,nnz+w_size);

		struct feature_node* wj = w_space;
		for(j=0;j<w_size;j++)
		{
			sparse_w_nnz[j] = jc[j+1] - jc[j];
			sparse_w[j] = wj;
			for(k=jc[j];k<jc[j+1];k++)
			{
				wj->index = (int) ir[k]+1;
				wj->value = ptr[k];
				wj++;
			}
			wj->index = -1;
			wj++;
		}

		if(w_size == 0) free(w_space);
	}
	else // ovr
	{
		ptr = mxGetPr(rhs[id]);
		model_->w=Malloc(double, w_size*nr_w);
		for(i = 0; i < w_size*nr_w; i++)
			model_->w[i]=ptr[i];
	}
	id++;

	// rho
	ptr = mxGetPr(rhs[id]);
	model_->rho = ptr[0];
	id++;

	mxFree(rhs);

	return NULL;
}

